/*
 * Copyright (c) 2010-2012 Marat Dukhan
 * Copyright (c) 2012 Georgia Institute of Technology
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are those
 * of the authors and should not be interpreted as representing official policies, 
 * either expressed or implied, of the Yeppp! Project.
 */

#pragma once

#if defined(_WIN32_WCE) || defined(WIN32_PLATFORM_HPC2000) || defined(WIN32_PLATFORM_HPCPRO) || defined(WIN32_PLATFORM_PSPC) || defined(WIN32_PLATFORM_WFSP)
	#define YEP_WINDOWSCE_OS
#endif
#if (defined(_WIN32) || defined(__WIN32__) || defined(__WINDOWS__)) && !defined(YEP_WINDOWSCE_PLATFORM)
	#define YEP_WINDOWS_OS
#endif
#if defined(__linux) || defined(__linux__) || defined(linux)
	#define YEP_LINUX_OS
#endif
#if defined(__APPLE__) && defined(__MACH__)
	#define YEP_MACOSX_OS
#endif

#if defined(_MSC_VER) && !defined(__INTEL_COMPILER)
	#define YEP_MICROSOFT_COMPILER
#endif
#if defined(__GNUC__) && !defined(__clang__) && !defined(__INTEL_COMPILER)
	#define YEP_GNU_COMPILER
#endif
#if defined(__INTEL_COMPILER)
	#define YEP_INTEL_COMPILER
	#if defined(_MSC_VER)
		#define YEP_INTEL_COMPILER_FOR_WINDOWS
	#else
		#define YEP_INTEL_COMPILER_FOR_UNIX
	#endif
#endif
#if defined(__clang__)
	#define YEP_CLANG_COMPILER
#endif
#if defined(__ARMCC_VERSION)
	#define YEP_ARM_COMPILER
#endif

#if defined(__pic__) || defined(__PIC__)
	#define YEP_PIC
#endif

#if defined(_VC_NODEFAULTLIB)
	#if !defined(YEP_CUSTOM_RUNTIME)
		#define YEP_CUSTOM_RUNTIME
	#endif
#endif

#define YEP_COUNT_OF(x) (sizeof(x) / sizeof(x[0]))

#if defined(_M_IX86) || defined(i386) || defined(__i386__) || defined(_X86_)
	#define YEP_X86_CPU
#endif
#if defined(_M_X64) || defined(_M_AMD64) || defined(__amd64__) || defined(__amd64) || defined(__x86_64__) || defined(__x86_64)
	#define YEP_X64_CPU
#endif
#if defined(_M_IA64) || defined(__itanium__) || defined(__ia64__) || defined(_IA64) || defined(__IA64__)
	#define YEP_IA64_CPU
#endif
#if defined(_M_ARM) || defined(__arm__)
	#define YEP_ARM_CPU
#endif
#if defined(_M_MRX000) || defined(_MIPS_) || defined(_MIPS64) || defined(__mips__) || defined(__mips) || defined(__MIPS__)
	#define YEP_MIPS_CPU
#endif
#if defined(__sparc__) || defined(__sparc)
	#define YEP_SPARC_CPU
#endif
#if defined(_M_PPC) || defined(__powerpc) || defined(__powerpc__) || defined(__POWERPC__) || defined(__ppc__)
	#define YEP_POWERPC_CPU
#endif

#if defined(YEP_X86_CPU)
	#if defined(YEP_MICROSOFT_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_WINDOWS)
		#if _M_IX86_FP >= 1
			#define YEP_CPU_SUPPORTS_X86_MMX_EXTENSION
			#define YEP_CPU_SUPPORTS_X86_MMXPLUS_EXTENSION
			#define YEP_CPU_SUPPORTS_X86_SSE_EXTENSION
		#endif
		#if _M_IX86_FP >= 2
			#define YEP_CPU_SUPPORTS_X86_SSE2_EXTENSION
		#endif
	#endif
	#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_UNIX)
		#if defined(__SSE__)
			#define YEP_CPU_SUPPORTS_X86_MMXPLUS_EXTENSION
			#define YEP_CPU_SUPPORTS_X86_SSE_EXTENSION
		#endif
		#if defined(__SSE2__)
			#define YEP_CPU_SUPPORTS_X86_SSE2_EXTENSION
		#endif
	#endif
#endif
#if defined(YEP_X64_CPU)
	#define YEP_X86_ISA_EXTENSION_SSE
	#define YEP_X86_ISA_EXTENSION_SSE2
#endif
#if defined(YEP_X86_CPU) || defined(YEP_X64_CPU)
	#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_UNIX)
		#if defined(__MMX__)
			#define YEP_CPU_SUPPORTS_X86_MMX_EXTENSION
		#endif
		#if defined(__3dNOW__)
			#define YEP_CPU_SUPPORTS_X86_3DNOW_EXTENSION
		#endif
		#if defined(__3dNOW_A__)
			#define YEP_CPU_SUPPORTS_X86_3DNOWPLUS_EXTENSION
		#endif
		#if defined(__SSE3__)
			#define YEP_CPU_SUPPORTS_X86_SSE3_EXTENSION
		#endif
		#if defined(__SSSE3__)
			#define YEP_CPU_SUPPORTS_X86_SSSE3_EXTENSION
		#endif
		#if defined(__SSE4A__)
			#define YEP_CPU_SUPPORTS_X86_SSE4A_EXTENSION
		#endif
		#if defined(__SSE4_1__)
			#define YEP_CPU_SUPPORTS_X86_SSE4_1_EXTENSION
		#endif
		#if defined(__SSE4_2__)
			#define YEP_CPU_SUPPORTS_X86_SSE4_2_EXTENSION
		#endif
		#if defined(__AVX__)
			#define YEP_CPU_SUPPORTS_X86_AVX_EXTENSION
		#endif
		#if defined(__F16C__)
			#define YEP_CPU_SUPPORTS_X86_F16C_EXTENSION
		#endif
		#if defined(__FMA4__)
			#define YEP_CPU_SUPPORTS_X86_FMA4_EXTENSION
			#define YEP_CPU_SUPPORTS_X86_FMA_EXTENSION
		#endif
		#if defined(__XOP__)
			#define YEP_CPU_SUPPORTS_X86_XOP_EXTENSION
		#endif
		#if defined(__ABM__)
			#define YEP_CPU_SUPPORTS_X86_LZCNT_EXTENSION
		#endif
		#if defined(__POPCNT__)
			#define YEP_CPU_SUPPORTS_X86_POPCNT_EXTENSION
		#endif
		#if defined(__BMI__)
			#define YEP_CPU_SUPPORTS_X86_BMI_EXTENSION
		#endif
		#if defined(__TBM__)
			#define YEP_CPU_SUPPORTS_X86_TBM_EXTENSION
		#endif
		#if defined(__MIC__)
			#define YEP_CPU_SUPPORTS_X86_KNF_EXTENSION
		#endif
		#if defined(__KNC__)
			#define YEP_CPU_SUPPORTS_X86_KNC_EXTENSION
		#endif
	#endif
#endif
#if defined(YEP_ARM_CPU)
	#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER)
		#if defined(__ARM_NEON__)
			#define YEP_CPU_SUPPORTS_ARM_NEON_EXTENSION
		#endif
	#endif
#endif
 
 #if defined(YEP_X86_CPU)
	#define YEP_X86_ABI
 #endif
 
#if defined(YEP_X64_CPU)
	#define YEP_X64_ABI
	#if defined(YEP_WINDOWS_OS)
		#define YEP_MICROSOFT_X64_ABI
	#else
		#define YEP_SYSTEMV_X64_ABI
	#endif
#endif

#if defined(YEP_INTEL_COMPILER) || defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER) || defined(YEP_ARM_COMPILER)
	/* These compilers support hex floats even in C++ */
	#define YEP_COMPILER_SUPPORTS_HEXADECIMAL_FLOATING_POINT_CONSTANTS
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ >= 199901L)
	/* C99 standard mandates support for hex floats */
	#define YEP_COMPILER_SUPPORTS_HEXADECIMAL_FLOATING_POINT_CONSTANTS
#endif

#if defined(__cplusplus) && (__cplusplus >= 201103L)
	#define YEP_NULL_POINTER nullptr
#else
	#define YEP_NULL_POINTER 0
#endif

#if defined(YEP_MICROSOFT_COMPILER)
	#define YEP_RESTRICT __restrict
#elif defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER)
	#define YEP_RESTRICT __restrict__
#elif defined(YEP_INTEL_COMPILER)
	#define YEP_RESTRICT restrict
#elif defined(YEP_ARM_COMPILER)
	#define YEP_RESTRICT __restrict__
#else
	#define YEP_RESTRICT
#endif

#if defined(YEP_MICROSOFT_COMPILER) || defined(YEP_INTEL_COMPILER)
	#define YEP_NOINLINE __declspec(noinline)
#elif defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER) || defined(YEP_ARM_COMPILER)
	#define YEP_NOINLINE __attribute__((noinline))
#else
	#define YEP_NOINLINE
#endif

#if defined(YEP_MICROSOFT_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_WINDOWS)
	#define YEP_INLINE __forceinline
#elif defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_UNIX)
	#define YEP_INLINE __attribute__((always_inline)) inline
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ >= 199901L)
	#define YEP_INLINE inline
#else
	#define YEP_INLINE
#endif

#if defined(__ELF__)
	#define YEP_PRIVATE_SYMBOL __attribute__((visibility ("internal")))
	#define YEP_LOCAL_SYMBOL   __attribute__((visibility ("hidden")))
	#define YEP_EXPORT_SYMBOL  __attribute__((visibility ("protected")))
	#define YEP_IMPORT_SYMBOL  __attribute__((visibility ("default")))
#elif defined(YEP_WINDOWS_OS)
	#define YEP_PRIVATE_SYMBOL
	#define YEP_LOCAL_SYMBOL
	#define YEP_EXPORT_SYMBOL __declspec(dllexport)
	#define YEP_IMPORT_SYMBOL __declspec(dllimport)
#else
	#define YEP_PRIVATE_SYMBOL
	#define YEP_LOCAL_SYMBOL
	#define YEP_EXPORT_SYMBOL
	#define YEP_IMPORT_SYMBOL
#endif
#define YEP_PUBLIC_SYMBOL YEP_IMPORT_SYMBOL

#if defined(YEP_X86_ABI) && defined(YEP_MICROSOFT_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_WINDOWS)
	#define YEPABI __cdecl
#elif defined(YEP_X86_ABI) && defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_UNIX)
	#define YEPABI __attribute__((cdecl))
#else
	#define YEPABI
#endif

#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_UNIX) || defined(YEP_ARM_COMPILER)
	/* Works with both data and code */
	#define YEP_USE_DATA_SECTION(name) __attribute__ ((section(name)))
#elif defined(YEP_MICROSOFT_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_WINDOWS)
	/* Works with data only */
	#define YEP_USE_DATA_SECTION(name) __declspec(allocate( name ))
#else
	#define YEP_USE_DATA_SECTION(name)
	#error Unsupported compiler
#endif

#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER)
	#define YEP_LIKELY(x) (__builtin_expect(!!(x), 1))
	#define YEP_UNLIKELY(x) (__builtin_expect(!!(x), 0))
#else
	#define YEP_LIKELY(x) (!!(x))
	#define YEP_UNLIKELY(x) (!!(x))
#endif

#ifdef __cplusplus
	#if defined(YEP_EXPORT_FUNCTIONS)
		#define YEP_DECLARE_FUNCTION(name, arguments) extern "C" YEP_PUBLIC_SYMBOL YepStatus YEPABI name arguments;
	#else
		#if defined(YEP_BUILD_LIBRARY)
			#define YEP_DECLARE_FUNCTION(name, arguments) extern "C" YEP_PUBLIC_SYMBOL YepStatus (YEPABI *name) arguments;
		#else
			#define YEP_DECLARE_FUNCTION(name, arguments) extern "C" YEP_PUBLIC_SYMBOL YepStatus (YEPABI *const name) arguments;
		#endif
	#endif
#else
	#if defined(YEP_EXPORT_FUNCTIONS)
		#define YEP_DECLARE_FUNCTION(name, arguments) extern YEP_PUBLIC_SYMBOL enum YepStatus YEPABI name arguments;
	#else
		#if defined(YEP_BUILD_LIBRARY)
			#define YEP_DECLARE_FUNCTION(name, arguments) extern YEP_PUBLIC_SYMBOL enum YepStatus (YEPABI *name) arguments;
		#else
			#define YEP_DECLARE_FUNCTION(name, arguments) extern YEP_PUBLIC_SYMBOL enum YepStatus (YEPABI *const name) arguments;
		#endif
	#endif
#endif

